"""GPS serial reader - parses NMEA sentences and dispatches to subscribers."""
from __future__ import annotations

import asyncio
import logging
import threading
from datetime import datetime, timezone
from typing import AsyncIterator, Callable

import pynmea2
import serial

from .exceptions import GPSDeviceNotFound, GPSPermissionError, GPSParseError
from .models import FixQuality, GPSData

logger = logging.getLogger(__name__)

_SUPPORTED = {"GGA", "RMC", "VTG"}

GPSCallback = Callable[[GPSData], None]


class GPSReader:
    """Reads data from the GPS device and dispatches to subscribers.

    Synchronous usage:
        reader = GPSReader()
        reader.add_callback(lambda d: print(d))
        reader.start()
        ...
        reader.stop()

    Asynchronous usage:
        async for data in GPSReader().stream():
            print(data)
    """

    DEFAULT_PORT = "/dev/ttyACM0"
    DEFAULT_BAUDRATE = 9600

    def __init__(
        self,
        port: str = DEFAULT_PORT,
        baudrate: int = DEFAULT_BAUDRATE,
        timeout: float = 2.0,
    ) -> None:
        self.port = port
        self.baudrate = baudrate
        self.timeout = timeout

        self._callbacks: list[GPSCallback] = []
        self._queues: list[asyncio.Queue[GPSData]] = []
        self._serial: serial.Serial | None = None
        self._thread: threading.Thread | None = None
        self._stop_event = threading.Event()
        self._loop: asyncio.AbstractEventLoop | None = None
        self._last_data = GPSData(timestamp=datetime.now(timezone.utc))

    # ------------------------------------------------------------------
    # Subscription API
    # ------------------------------------------------------------------

    def add_callback(self, cb: GPSCallback) -> None:
        """Register a callback to be called on every new GPS reading."""
        self._callbacks.append(cb)

    def remove_callback(self, cb: GPSCallback) -> None:
        self._callbacks.remove(cb)

    def _make_queue(self, maxsize: int = 100) -> asyncio.Queue[GPSData]:
        q: asyncio.Queue[GPSData] = asyncio.Queue(maxsize=maxsize)
        self._queues.append(q)
        return q

    # ------------------------------------------------------------------
    # Start / Stop
    # ------------------------------------------------------------------

    def start(self) -> None:
        """Start the background reading thread."""
        if self._thread and self._thread.is_alive():
            return
        self._stop_event.clear()
        try:
            self._loop = asyncio.get_event_loop()
        except RuntimeError:
            self._loop = None
        self._thread = threading.Thread(target=self._run, daemon=True, name="gps-reader")
        self._thread.start()
        logger.info("GPSReader started: %s @ %d baud", self.port, self.baudrate)

    def stop(self) -> None:
        """Stop reading."""
        self._stop_event.set()
        if self._thread:
            self._thread.join(timeout=5)
        if self._serial and self._serial.is_open:
            self._serial.close()
        logger.info("GPSReader stopped.")

    # ------------------------------------------------------------------
    # Async stream interface
    # ------------------------------------------------------------------

    async def stream(self, maxsize: int = 100) -> AsyncIterator[GPSData]:
        """Async generator: use as `async for data in reader.stream()`."""
        self._loop = asyncio.get_running_loop()
        q = self._make_queue(maxsize)
        self.start()
        try:
            while True:
                data = await q.get()
                yield data
        finally:
            self._queues.remove(q)

    # ------------------------------------------------------------------
    # Internal read loop
    # ------------------------------------------------------------------

    def _open_serial(self) -> serial.Serial:
        try:
            return serial.Serial(self.port, self.baudrate, timeout=self.timeout)
        except serial.SerialException as e:
            msg = str(e).lower()
            if "permission" in msg or "access" in msg:
                raise GPSPermissionError(
                    f"No permission to access {self.port}. "
                    "Fix: sudo usermod -aG dialout $USER && (re-login)"
                ) from e
            raise GPSDeviceNotFound(f"Could not open port: {self.port} — {e}") from e

    def _run(self) -> None:
        try:
            self._serial = self._open_serial()
        except Exception as e:
            logger.error("Failed to open serial port: %s", e)
            return

        partial: dict = {}  # accumulates GGA, RMC, VTG messages

        while not self._stop_event.is_set():
            try:
                raw = self._serial.readline()
                if not raw:
                    continue
                line = raw.decode("ascii", errors="ignore").strip()
                if not line.startswith("$"):
                    continue

                data = self._parse_line(line, partial)
                if data is not None:
                    self._last_data = data
                    self._dispatch(data)

            except serial.SerialException as e:
                logger.error("Serial read error: %s", e)
                break
            except Exception as e:
                logger.debug("Line processing error: %s", e)

    def _parse_line(self, line: str, partial: dict) -> GPSData | None:
        """Parse an NMEA line; returns GPSData when a full reading is ready."""
        try:
            msg = pynmea2.parse(line)
        except pynmea2.ParseError:
            return None

        sentence_type = msg.sentence_type

        if sentence_type == "GGA":
            partial["gga"] = msg
        elif sentence_type == "RMC":
            partial["rmc"] = msg
        elif sentence_type == "VTG":
            partial["vtg"] = msg
        else:
            return None

        if "gga" not in partial:
            return None

        return self._build_data(partial)

    def _build_data(self, partial: dict) -> GPSData:
        gga = partial.get("gga")
        rmc = partial.get("rmc")
        vtg = partial.get("vtg")

        now = datetime.now(timezone.utc)

        # Timestamp
        ts = now
        if gga and gga.timestamp:
            try:
                ts = datetime.combine(now.date(), gga.timestamp, tzinfo=timezone.utc)
            except (TypeError, ValueError, OverflowError):
                pass

        # Position — pynmea2's .latitude/.longitude already apply N/S/E/W sign
        lat = lon = None
        try:
            if gga and gga.lat:
                lat = gga.latitude
            if gga and gga.lon:
                lon = gga.longitude
        except (ValueError, AttributeError):
            pass

        # Altitude
        alt = None
        try:
            if gga and gga.altitude:
                alt = float(gga.altitude)
        except (ValueError, AttributeError):
            pass

        # Speed (VTG km/h → m/s, fallback to RMC knots → m/s)
        speed = None
        try:
            if vtg and vtg.spd_over_grnd_kmph:
                speed = float(vtg.spd_over_grnd_kmph) / 3.6
            elif rmc and rmc.spd_over_grnd:
                speed = float(rmc.spd_over_grnd) * 0.514444  # knots → m/s
            # Noise filter: treat speeds below 1.5 km/h as stationary
            if speed is not None and speed < (1.5 / 3.6):
                speed = 0.0
        except (ValueError, AttributeError):
            pass

        # Heading
        heading = None
        try:
            if vtg and vtg.true_track:
                heading = float(vtg.true_track)
            elif rmc and rmc.true_course:
                heading = float(rmc.true_course)
        except (ValueError, AttributeError):
            pass

        # Fix quality
        fix_q = FixQuality.NO_FIX
        try:
            if gga and gga.gps_qual is not None:
                fix_q = FixQuality(int(gga.gps_qual))
        except (ValueError, AttributeError):
            pass

        # Satellite count
        num_sats = 0
        try:
            if gga and gga.num_sats:
                num_sats = int(gga.num_sats)
        except (ValueError, AttributeError):
            pass

        # HDOP
        hdop = None
        try:
            if gga and gga.horizontal_dil:
                hdop = float(gga.horizontal_dil)
        except (ValueError, AttributeError):
            pass

        return GPSData(
            timestamp=ts,
            latitude=lat,
            longitude=lon,
            altitude=alt,
            speed=speed,
            heading=heading,
            fix_quality=fix_q,
            num_satellites=num_sats,
            hdop=hdop,
            raw_sentence=gga.render() if gga else "",
        )

    # ------------------------------------------------------------------
    # Dispatch
    # ------------------------------------------------------------------

    def _dispatch(self, data: GPSData) -> None:
        # Synchronous callbacks
        for cb in list(self._callbacks):
            try:
                cb(data)
            except Exception as e:
                logger.warning("Callback error: %s", e)

        # Async queues
        if self._loop and self._loop.is_running() and self._queues:
            for q in list(self._queues):
                try:
                    self._loop.call_soon_threadsafe(q.put_nowait, data)
                except asyncio.QueueFull:
                    logger.debug("Queue full, data dropped.")

    # ------------------------------------------------------------------
    # Status
    # ------------------------------------------------------------------

    @property
    def last(self) -> GPSData:
        """Most recently received GPS data."""
        return self._last_data

    @property
    def is_running(self) -> bool:
        return bool(self._thread and self._thread.is_alive())

    def __enter__(self) -> "GPSReader":
        self.start()
        return self

    def __exit__(self, *_) -> None:
        self.stop()



"""Tests for gps_mouse.reader — NMEA parsing and noise filter."""
import time
from datetime import datetime, timezone
from unittest.mock import MagicMock, patch

import pytest
import serial

from gps_mouse.exceptions import GPSDeviceNotFound, GPSPermissionError
from gps_mouse.models import FixQuality, GPSData
from gps_mouse.reader import GPSReader


# ------------------------------------------------------------------
# Valid NMEA sentences (checksums verified)
# ------------------------------------------------------------------

GGA_FIX     = "$GPGGA,120000.00,3958.3603,N,03238.4902,E,1,07,1.20,858.5,M,32.1,M,,*5F"
GGA_NO_FIX  = "$GPGGA,120000.00,,,,,,0,00,99.99,,M,,M,,*49"
RMC_MOVING  = "$GPRMC,120000.00,A,3958.3603,N,03238.4902,E,5.00,90.0,250326,,,A*55"
VTG_MOVING  = "$GPVTG,90.0,T,,M,9.26,N,17.16,K,A*08"
VTG_STILL   = "$GPVTG,,,,,0.00,N,0.00,K,N*35"


def make_reader(**kwargs) -> GPSReader:
    defaults = {"port": "/dev/ttyACM0"}
    defaults.update(kwargs)
    return GPSReader(**defaults)


# ------------------------------------------------------------------
# _parse_line
# ------------------------------------------------------------------

class TestParseLine:
    def test_gga_stored_in_partial(self):
        reader = make_reader()
        partial = {}
        result = reader._parse_line(GGA_FIX, partial)
        assert "gga" in partial
        assert result is not None

    def test_rmc_stored_but_no_output_without_gga(self):
        reader = make_reader()
        partial = {}
        result = reader._parse_line(RMC_MOVING, partial)
        assert "rmc" in partial
        assert result is None  # no GGA yet

    def test_vtg_stored_but_no_output_without_gga(self):
        reader = make_reader()
        partial = {}
        result = reader._parse_line(VTG_MOVING, partial)
        assert "vtg" in partial
        assert result is None

    def test_gga_after_rmc_produces_data(self):
        reader = make_reader()
        partial = {}
        reader._parse_line(RMC_MOVING, partial)
        result = reader._parse_line(GGA_FIX, partial)
        assert result is not None
        assert isinstance(result, GPSData)

    def test_invalid_sentence_returns_none(self):
        reader = make_reader()
        result = reader._parse_line("$GPXXX,garbage,,*00", {})
        assert result is None

    def test_non_nmea_line_returns_none(self):
        reader = make_reader()
        result = reader._parse_line("not nmea at all", {})
        assert result is None

    def test_unknown_sentence_type_returns_none(self):
        reader = make_reader()
        partial = {}
        result = reader._parse_line("$GPGSV,3,1,12*73", partial)
        assert result is None


# ------------------------------------------------------------------
# _build_data — position
# ------------------------------------------------------------------

class TestBuildDataPosition:
    def _build(self, gga_str, rmc_str=None, vtg_str=None) -> GPSData:
        reader = make_reader()
        partial = {}
        reader._parse_line(gga_str, partial)
        if rmc_str:
            reader._parse_line(rmc_str, partial)
        if vtg_str:
            reader._parse_line(vtg_str, partial)
        return reader._build_data(partial)

    def test_latitude_north(self):
        data = self._build(GGA_FIX)
        assert data.latitude == pytest.approx(39.972672, abs=1e-4)

    def test_longitude_east(self):
        data = self._build(GGA_FIX)
        assert data.longitude == pytest.approx(32.641503, abs=1e-4)

    def test_altitude(self):
        data = self._build(GGA_FIX)
        assert data.altitude == pytest.approx(858.5, abs=0.1)

    def test_fix_quality_gps(self):
        data = self._build(GGA_FIX)
        assert data.fix_quality == FixQuality.GPS

    def test_num_satellites(self):
        data = self._build(GGA_FIX)
        assert data.num_satellites == 7

    def test_hdop(self):
        data = self._build(GGA_FIX)
        assert data.hdop == pytest.approx(1.2, abs=0.01)

    def test_no_fix_quality(self):
        data = self._build(GGA_NO_FIX)
        assert data.fix_quality == FixQuality.NO_FIX

    def test_no_fix_latitude_none(self):
        data = self._build(GGA_NO_FIX)
        assert data.latitude is None

    def test_southern_hemisphere(self):
        # Recompute checksum after replacing N → S
        body = "GPGGA,120000.00,3958.3603,S,03238.4902,E,1,07,1.20,858.5,M,32.1,M,,"
        cs = 0
        for c in body:
            cs ^= ord(c)
        gga_south = f"${body}*{cs:02X}"
        data = self._build(gga_south)
        assert data.latitude < 0

    def test_western_hemisphere(self):
        body = "GPGGA,120000.00,3958.3603,N,03238.4902,W,1,07,1.20,858.5,M,32.1,M,,"
        cs = 0
        for c in body:
            cs ^= ord(c)
        gga_west = f"${body}*{cs:02X}"
        data = self._build(gga_west)
        assert data.longitude < 0


# ------------------------------------------------------------------
# _build_data — speed noise filter
# ------------------------------------------------------------------

class TestSpeedNoise:
    def _build_with_vtg_kmh(self, kmh: float) -> GPSData:
        reader = make_reader()
        partial = {}
        reader._parse_line(GGA_FIX, partial)
        vtg_mock = MagicMock()
        vtg_mock.sentence_type = "VTG"
        vtg_mock.spd_over_grnd_kmph = str(kmh)
        vtg_mock.true_track = "90.0"
        partial["vtg"] = vtg_mock
        return reader._build_data(partial)

    def test_speed_above_threshold_passes(self):
        data = self._build_with_vtg_kmh(5.0)
        assert data.speed is not None
        assert data.speed > 0

    def test_speed_below_threshold_zeroed(self):
        data = self._build_with_vtg_kmh(1.0)
        assert data.speed == 0.0

    def test_speed_at_noise_floor_zeroed(self):
        data = self._build_with_vtg_kmh(1.4)
        assert data.speed == 0.0

    def test_speed_kmh_converts_correctly(self):
        data = self._build_with_vtg_kmh(36.0)
        assert data.speed == pytest.approx(10.0, abs=0.01)
        assert data.speed_kmh == pytest.approx(36.0, abs=0.1)


# ------------------------------------------------------------------
# _open_serial — permission / not-found errors
# ------------------------------------------------------------------

class TestOpenSerial:
    def test_permission_error_raises_gps_permission_error(self):
        reader = make_reader()
        with patch("serial.Serial", side_effect=serial.SerialException("permission denied")):
            with pytest.raises(GPSPermissionError):
                reader._open_serial()

    def test_not_found_raises_gps_device_not_found(self):
        reader = make_reader(port="/dev/nonexistent")
        with patch("serial.Serial", side_effect=serial.SerialException("no such file")):
            with pytest.raises(GPSDeviceNotFound):
                reader._open_serial()


# ------------------------------------------------------------------
# Callback dispatch
# ------------------------------------------------------------------

class TestCallbacks:
    def test_callback_called_on_data(self):
        received = []
        reader = make_reader()
        reader.add_callback(received.append)

        partial = {}
        reader._parse_line(GGA_FIX, partial)
        data = reader._build_data(partial)
        reader._dispatch(data)

        assert len(received) == 1
        assert received[0].has_fix

    def test_multiple_callbacks(self):
        results = [[], []]
        reader = make_reader()
        reader.add_callback(results[0].append)
        reader.add_callback(results[1].append)

        partial = {}
        reader._parse_line(GGA_FIX, partial)
        data = reader._build_data(partial)
        reader._dispatch(data)

        assert len(results[0]) == 1
        assert len(results[1]) == 1

    def test_failing_callback_does_not_stop_others(self):
        received = []
        reader = make_reader()
        reader.add_callback(lambda d: (_ for _ in ()).throw(RuntimeError("boom")))
        reader.add_callback(received.append)

        partial = {}
        reader._parse_line(GGA_FIX, partial)
        data = reader._build_data(partial)
        reader._dispatch(data)

        assert len(received) == 1

    def test_remove_callback(self):
        received = []
        reader = make_reader()
        cb = received.append
        reader.add_callback(cb)
        reader.remove_callback(cb)

        partial = {}
        reader._parse_line(GGA_FIX, partial)
        data = reader._build_data(partial)
        reader._dispatch(data)

        assert received == []


# ------------------------------------------------------------------
# Start / Stop / is_running
# ------------------------------------------------------------------

class TestStartStop:
    def test_not_running_initially(self):
        reader = make_reader()
        assert reader.is_running is False

    def test_start_fails_gracefully_when_port_missing(self):
        reader = make_reader(port="/dev/nonexistent999")
        reader.start()
        time.sleep(0.3)
        reader.stop()

    def test_context_manager_stops_reader(self):
        reader = make_reader(port="/dev/nonexistent999")
        with reader:
            pass
        assert reader.is_running is False

    def test_double_start_is_safe(self):
        reader = make_reader(port="/dev/nonexistent999")
        reader.start()
        reader.start()  # second call should be a no-op
        reader.stop()

    def test_last_returns_gpsdata(self):
        reader = make_reader()
        assert isinstance(reader.last, GPSData)



"""GPS library exceptions."""


class GPSError(Exception):
    """Base GPS error."""


class GPSDeviceNotFound(GPSError):
    """Serial port not found or could not be opened."""


class GPSPermissionError(GPSError):
    """No permission to access the serial port.

    Fix: sudo usermod -aG dialout $USER
    """


class GPSParseError(GPSError):
    """Failed to parse NMEA sentence."""


class GPSNotRunning(GPSError):
    """Reader has not been started yet."""


class GPSPublisherError(GPSError):
    """ZMQ publish error."""


class GPSSubscriberError(GPSError):
    """ZMQ subscribe error."""



"""ZMQ-based GPS data publisher.

Other processes can connect using GPSSubscriber to receive data.

Architecture:
    GPSReader → GPSPublisher (PUB socket) ← GPSSubscriber (SUB socket)

Default address: tcp://127.0.0.1:5557
"""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import zmq

from .exceptions import GPSPublisherError
from .models import GPSData

if TYPE_CHECKING:
    from .reader import GPSReader

logger = logging.getLogger(__name__)

DEFAULT_ADDRESS = "tcp://127.0.0.1:5557"


class GPSPublisher:
    """Broadcasts GPS data over a ZMQ PUB socket.

    Usage:
        reader = GPSReader()
        pub = GPSPublisher()
        pub.attach(reader)   # registers as a reader callback automatically
        reader.start()
    """

    def __init__(self, address: str = DEFAULT_ADDRESS) -> None:
        self.address = address
        self._ctx = zmq.Context.instance()
        self._socket: zmq.Socket | None = None

    # ------------------------------------------------------------------
    # Lifecycle
    # ------------------------------------------------------------------

    def start(self) -> None:
        if self._socket is not None:
            return
        try:
            self._socket = self._ctx.socket(zmq.PUB)
            self._socket.bind(self.address)
            logger.info("GPSPublisher listening: %s", self.address)
        except zmq.ZMQError as e:
            raise GPSPublisherError(f"ZMQ bind error {self.address}: {e}") from e

    def stop(self) -> None:
        if self._socket:
            self._socket.close()
            self._socket = None
        logger.info("GPSPublisher closed.")

    # ------------------------------------------------------------------
    # Attach to reader
    # ------------------------------------------------------------------

    def attach(self, reader: "GPSReader") -> None:
        """Can be called before the reader is started."""
        self.start()
        reader.add_callback(self.publish)

    def detach(self, reader: "GPSReader") -> None:
        reader.remove_callback(self.publish)

    # ------------------------------------------------------------------
    # Publish
    # ------------------------------------------------------------------

    def publish(self, data: GPSData) -> None:
        """Publish a GPSData object as JSON."""
        if self._socket is None:
            self.start()
        try:
            topic = "gps.fix" if data.has_fix else "gps.nofix"
            payload = f"{topic} {data.to_json()}"
            self._socket.send_string(payload, zmq.NOBLOCK)
        except zmq.Again:
            pass  # no subscribers, ignore
        except zmq.ZMQError as e:
            logger.warning("Publish error: %s", e)

    # ------------------------------------------------------------------
    # Context manager
    # ------------------------------------------------------------------

    def __enter__(self) -> "GPSPublisher":
        self.start()
        return self

    def __exit__(self, *_) -> None:
        self.stop()



"""Tests for gps_mouse.subscriber."""
import json
from datetime import datetime, timezone

import pytest

from gps_mouse.models import FixQuality, GPSData
from gps_mouse.subscriber import GPSSubscriber


TS = datetime(2026, 3, 25, 12, 0, 0, tzinfo=timezone.utc)


def make_fix_json() -> str:
    data = GPSData(
        timestamp=TS,
        latitude=39.972672,
        longitude=32.641504,
        altitude=858.5,
        speed=0.0,
        fix_quality=FixQuality.GPS,
        num_satellites=7,
        hdop=1.2,
    )
    return data.to_json()


def make_zmq_message(topic: str, data_json: str) -> str:
    return f"{topic} {data_json}"


# ------------------------------------------------------------------
# _parse_message
# ------------------------------------------------------------------

class TestParseMessage:
    def test_valid_fix_message(self):
        msg = make_zmq_message("gps.fix", make_fix_json())
        data = GPSSubscriber._parse_message(msg)
        assert data is not None
        assert data.has_fix
        assert data.latitude == pytest.approx(39.972672)

    def test_valid_nofix_message(self):
        nofix = GPSData(timestamp=TS).to_json()
        msg = make_zmq_message("gps.nofix", nofix)
        data = GPSSubscriber._parse_message(msg)
        assert data is not None
        assert data.has_fix is False

    def test_malformed_message_returns_none(self):
        data = GPSSubscriber._parse_message("no-space-here")
        assert data is None

    def test_invalid_json_returns_none(self):
        data = GPSSubscriber._parse_message("gps.fix {not valid json}")
        assert data is None

    def test_empty_string_returns_none(self):
        data = GPSSubscriber._parse_message("")
        assert data is None

    def test_latitude_preserved(self):
        msg = make_zmq_message("gps.fix", make_fix_json())
        data = GPSSubscriber._parse_message(msg)
        assert data.latitude == pytest.approx(39.972672)

    def test_longitude_preserved(self):
        msg = make_zmq_message("gps.fix", make_fix_json())
        data = GPSSubscriber._parse_message(msg)
        assert data.longitude == pytest.approx(32.641504)

    def test_altitude_preserved(self):
        msg = make_zmq_message("gps.fix", make_fix_json())
        data = GPSSubscriber._parse_message(msg)
        assert data.altitude == pytest.approx(858.5)

    def test_num_satellites_preserved(self):
        msg = make_zmq_message("gps.fix", make_fix_json())
        data = GPSSubscriber._parse_message(msg)
        assert data.num_satellites == 7

    def test_fix_quality_is_enum(self):
        msg = make_zmq_message("gps.fix", make_fix_json())
        data = GPSSubscriber._parse_message(msg)
        assert isinstance(data.fix_quality, FixQuality)

    def test_timestamp_is_datetime(self):
        msg = make_zmq_message("gps.fix", make_fix_json())
        data = GPSSubscriber._parse_message(msg)
        assert isinstance(data.timestamp, datetime)

    def test_json_with_extra_spaces_in_topic(self):
        # topic and json separated by single space — standard
        msg = "gps.fix " + make_fix_json()
        data = GPSSubscriber._parse_message(msg)
        assert data is not None


# ------------------------------------------------------------------
# Constructor defaults
# ------------------------------------------------------------------

class TestConstructor:
    def test_default_address(self):
        from gps_mouse.publisher import DEFAULT_ADDRESS
        sub = GPSSubscriber()
        assert sub.address == DEFAULT_ADDRESS

    def test_default_topic(self):
        sub = GPSSubscriber()
        assert sub.topic == "gps"

    def test_custom_address(self):
        sub = GPSSubscriber(address="tcp://192.168.1.10:5557")
        assert sub.address == "tcp://192.168.1.10:5557"

    def test_custom_topic(self):
        sub = GPSSubscriber(topic="gps.fix")
        assert sub.topic == "gps.fix"



"""Tests for gps_mouse.publisher."""
from datetime import datetime, timezone
from unittest.mock import MagicMock, patch

import pytest
import zmq

from gps_mouse.exceptions import GPSPublisherError
from gps_mouse.models import FixQuality, GPSData
from gps_mouse.publisher import GPSPublisher, DEFAULT_ADDRESS


TS = datetime(2026, 3, 25, 12, 0, 0, tzinfo=timezone.utc)


def make_data(has_fix=True) -> GPSData:
    if has_fix:
        return GPSData(
            timestamp=TS,
            latitude=39.972672,
            longitude=32.641504,
            altitude=858.5,
            fix_quality=FixQuality.GPS,
            num_satellites=7,
        )
    return GPSData(timestamp=TS)


# ------------------------------------------------------------------
# Default address
# ------------------------------------------------------------------

def test_default_address():
    pub = GPSPublisher()
    assert pub.address == DEFAULT_ADDRESS


def test_custom_address():
    pub = GPSPublisher(address="tcp://127.0.0.1:9999")
    assert pub.address == "tcp://127.0.0.1:9999"


# ------------------------------------------------------------------
# start / stop
# ------------------------------------------------------------------

class TestLifecycle:
    def test_start_binds_socket(self):
        mock_sock = MagicMock()
        mock_ctx = MagicMock()
        mock_ctx.socket.return_value = mock_sock

        with patch("zmq.Context.instance", return_value=mock_ctx):
            pub = GPSPublisher(address="tcp://127.0.0.1:19001")
            pub.start()
            mock_sock.bind.assert_called_once_with("tcp://127.0.0.1:19001")

    def test_double_start_binds_only_once(self):
        mock_sock = MagicMock()
        mock_ctx = MagicMock()
        mock_ctx.socket.return_value = mock_sock

        with patch("zmq.Context.instance", return_value=mock_ctx):
            pub = GPSPublisher(address="tcp://127.0.0.1:19002")
            pub.start()
            pub.start()
            assert mock_sock.bind.call_count == 1

    def test_stop_closes_socket(self):
        mock_sock = MagicMock()
        mock_ctx = MagicMock()
        mock_ctx.socket.return_value = mock_sock

        with patch("zmq.Context.instance", return_value=mock_ctx):
            pub = GPSPublisher(address="tcp://127.0.0.1:19003")
            pub.start()
            pub.stop()
            mock_sock.close.assert_called_once()

    def test_start_zmq_error_raises_publisher_error(self):
        mock_sock = MagicMock()
        mock_sock.bind.side_effect = zmq.ZMQError("address in use")
        mock_ctx = MagicMock()
        mock_ctx.socket.return_value = mock_sock

        with patch("zmq.Context.instance", return_value=mock_ctx):
            pub = GPSPublisher(address="tcp://127.0.0.1:19004")
            with pytest.raises(GPSPublisherError):
                pub.start()

    def test_context_manager(self):
        mock_sock = MagicMock()
        mock_ctx = MagicMock()
        mock_ctx.socket.return_value = mock_sock

        with patch("zmq.Context.instance", return_value=mock_ctx):
            with GPSPublisher(address="tcp://127.0.0.1:19005") as pub:
                pass
            mock_sock.close.assert_called_once()


# ------------------------------------------------------------------
# publish — topic selection
# ------------------------------------------------------------------

class TestPublish:
    def _make_pub_with_mock(self, address="tcp://127.0.0.1:19010"):
        mock_sock = MagicMock()
        mock_ctx = MagicMock()
        mock_ctx.socket.return_value = mock_sock
        with patch("zmq.Context.instance", return_value=mock_ctx):
            pub = GPSPublisher(address=address)
            pub.start()
            return pub, mock_sock

    def test_fix_uses_gps_fix_topic(self):
        pub, sock = self._make_pub_with_mock("tcp://127.0.0.1:19011")
        pub.publish(make_data(has_fix=True))
        call_args = sock.send_string.call_args[0][0]
        assert call_args.startswith("gps.fix ")

    def test_no_fix_uses_gps_nofix_topic(self):
        pub, sock = self._make_pub_with_mock("tcp://127.0.0.1:19012")
        pub.publish(make_data(has_fix=False))
        call_args = sock.send_string.call_args[0][0]
        assert call_args.startswith("gps.nofix ")

    def test_payload_is_valid_json(self):
        import json
        pub, sock = self._make_pub_with_mock("tcp://127.0.0.1:19013")
        pub.publish(make_data(has_fix=True))
        raw = sock.send_string.call_args[0][0]
        _, json_part = raw.split(" ", 1)
        parsed = json.loads(json_part)
        assert "latitude" in parsed
        assert parsed["latitude"] == pytest.approx(39.972672)

    def test_zmq_again_is_silently_ignored(self):
        pub, sock = self._make_pub_with_mock("tcp://127.0.0.1:19014")
        sock.send_string.side_effect = zmq.Again()
        pub.publish(make_data())  # must not raise

    def test_attach_registers_callback(self):
        from gps_mouse.reader import GPSReader
        mock_sock = MagicMock()
        mock_ctx = MagicMock()
        mock_ctx.socket.return_value = mock_sock

        with patch("zmq.Context.instance", return_value=mock_ctx):
            pub = GPSPublisher(address="tcp://127.0.0.1:19015")
            reader = GPSReader(port="/dev/nonexistent")
            pub.attach(reader)
            assert pub.publish in reader._callbacks

    def test_detach_removes_callback(self):
        from gps_mouse.reader import GPSReader
        mock_sock = MagicMock()
        mock_ctx = MagicMock()
        mock_ctx.socket.return_value = mock_sock

        with patch("zmq.Context.instance", return_value=mock_ctx):
            pub = GPSPublisher(address="tcp://127.0.0.1:19016")
            reader = GPSReader(port="/dev/nonexistent")
            pub.attach(reader)
            pub.detach(reader)
            assert pub.publish not in reader._callbacks



"""gps_mouse — U-Blox GPS library.

Quick start:

    # Start GPS reader + ZMQ publisher (main project)
    from gps_mouse import GPSReader, GPSPublisher

    reader = GPSReader()
    pub = GPSPublisher()
    pub.attach(reader)
    reader.start()

    # Subscribe from another project
    from gps_mouse import GPSSubscriber

    for data in GPSSubscriber().iter_fixes():
        print(data.latitude, data.longitude)
"""

from .exceptions import (
    GPSDeviceNotFound,
    GPSError,
    GPSNotRunning,
    GPSParseError,
    GPSPermissionError,
    GPSPublisherError,
    GPSSubscriberError,
)
from .models import FixQuality, GPSData
from .publisher import DEFAULT_ADDRESS, GPSPublisher
from .reader import GPSReader
from .subscriber import GPSSubscriber

__all__ = [
    # Main classes
    "GPSReader",
    "GPSPublisher",
    "GPSSubscriber",
    # Data model
    "GPSData",
    "FixQuality",
    # Constants
    "DEFAULT_ADDRESS",
    # Exceptions
    "GPSError",
    "GPSDeviceNotFound",
    "GPSPermissionError",
    "GPSParseError",
    "GPSNotRunning",
    "GPSPublisherError",
    "GPSSubscriberError",
]

__version__ = "0.1.0"



"""ZMQ GPS subscriber - for other projects to receive GPS data.

Synchronous usage:
    sub = GPSSubscriber()
    for data in sub.iter_fixes():
        print(data.latitude, data.longitude)

Asynchronous usage:
    async for data in GPSSubscriber().stream():
        print(data)

One-shot usage:
    data = GPSSubscriber().wait_for_fix(timeout=10)
"""
from __future__ import annotations

import asyncio
import logging
from typing import AsyncIterator, Iterator

import zmq
import zmq.asyncio

from .exceptions import GPSSubscriberError
from .models import GPSData
from .publisher import DEFAULT_ADDRESS

logger = logging.getLogger(__name__)


class GPSSubscriber:
    """Receives GPS data over ZMQ.

    Args:
        address: Publisher address (default tcp://127.0.0.1:5557)
        topic:   Filter ("gps.fix", "gps.nofix", "gps" → all)
    """

    def __init__(
        self,
        address: str = DEFAULT_ADDRESS,
        topic: str = "gps",
    ) -> None:
        self.address = address
        self.topic = topic

    # ------------------------------------------------------------------
    # Synchronous interface
    # ------------------------------------------------------------------

    def iter_fixes(self, timeout_ms: int = 5000) -> Iterator[GPSData]:
        """Synchronous generator: yields GPSData for every incoming reading."""
        ctx = zmq.Context.instance()
        sock = ctx.socket(zmq.SUB)
        sock.connect(self.address)
        sock.setsockopt_string(zmq.SUBSCRIBE, self.topic)
        sock.setsockopt(zmq.RCVTIMEO, timeout_ms)
        logger.info("GPSSubscriber connected: %s (topic=%s)", self.address, self.topic)
        try:
            while True:
                try:
                    msg = sock.recv_string()
                    data = self._parse_message(msg)
                    if data:
                        yield data
                except zmq.Again:
                    logger.debug("Timeout — waiting for data…")
        finally:
            sock.close()

    def wait_for_fix(self, timeout: float = 10.0) -> GPSData | None:
        """Wait for the first fix and return it. Returns None on timeout."""
        ctx = zmq.Context.instance()
        sock = ctx.socket(zmq.SUB)
        sock.connect(self.address)
        sock.setsockopt_string(zmq.SUBSCRIBE, "gps.fix")
        sock.setsockopt(zmq.RCVTIMEO, int(timeout * 1000))
        try:
            msg = sock.recv_string()
            return self._parse_message(msg)
        except zmq.Again:
            return None
        finally:
            sock.close()

    # ------------------------------------------------------------------
    # Asynchronous interface
    # ------------------------------------------------------------------

    async def stream(self) -> AsyncIterator[GPSData]:
        """Async generator: use as `async for data in sub.stream()`."""
        ctx = zmq.asyncio.Context.instance()
        sock = ctx.socket(zmq.SUB)
        sock.connect(self.address)
        sock.setsockopt_string(zmq.SUBSCRIBE, self.topic)
        logger.info("GPSSubscriber (async) connected: %s", self.address)
        try:
            while True:
                try:
                    msg = await asyncio.wait_for(sock.recv_string(), timeout=5.0)
                    data = self._parse_message(msg)
                    if data:
                        yield data
                except asyncio.TimeoutError:
                    logger.debug("Async timeout — waiting…")
        finally:
            sock.close()

    async def wait_for_fix_async(self, timeout: float = 10.0) -> GPSData | None:
        """Async: wait for the first fix and return it."""
        ctx = zmq.asyncio.Context.instance()
        sock = ctx.socket(zmq.SUB)
        sock.connect(self.address)
        sock.setsockopt_string(zmq.SUBSCRIBE, "gps.fix")
        try:
            msg = await asyncio.wait_for(sock.recv_string(), timeout=timeout)
            return self._parse_message(msg)
        except asyncio.TimeoutError:
            return None
        finally:
            sock.close()

    # ------------------------------------------------------------------
    # Helpers
    # ------------------------------------------------------------------

    @staticmethod
    def _parse_message(msg: str) -> GPSData | None:
        try:
            _, json_part = msg.split(" ", 1)
            return GPSData.from_json(json_part)
        except Exception as e:
            logger.warning("Failed to parse message: %s — %s", msg[:60], e)
            return None



"""Tests for gps_mouse.models."""
import json
from datetime import datetime, timezone

import pytest

from gps_mouse.models import FixQuality, GPSData


TS = datetime(2026, 3, 25, 12, 0, 0, tzinfo=timezone.utc)


def make_fix(**kwargs) -> GPSData:
    defaults = dict(
        timestamp=TS,
        latitude=39.972672,
        longitude=32.641504,
        altitude=858.5,
        speed=2.0,
        heading=90.0,
        fix_quality=FixQuality.GPS,
        num_satellites=7,
        hdop=1.2,
    )
    defaults.update(kwargs)
    return GPSData(**defaults)


# ------------------------------------------------------------------
# FixQuality
# ------------------------------------------------------------------

class TestFixQuality:
    def test_values(self):
        assert FixQuality.NO_FIX == 0
        assert FixQuality.GPS == 1
        assert FixQuality.DGPS == 2
        assert FixQuality.RTK == 4

    def test_ordering(self):
        assert FixQuality.NO_FIX < FixQuality.GPS
        assert FixQuality.GPS < FixQuality.RTK


# ------------------------------------------------------------------
# GPSData.has_fix
# ------------------------------------------------------------------

class TestHasFix:
    def test_has_fix_when_quality_and_lat_present(self):
        data = make_fix()
        assert data.has_fix is True

    def test_no_fix_when_quality_zero(self):
        data = make_fix(fix_quality=FixQuality.NO_FIX)
        assert data.has_fix is False

    def test_no_fix_when_latitude_none(self):
        data = make_fix(latitude=None)
        assert data.has_fix is False

    def test_no_fix_when_both_missing(self):
        data = GPSData(timestamp=TS)
        assert data.has_fix is False


# ------------------------------------------------------------------
# GPSData.speed_kmh
# ------------------------------------------------------------------

class TestSpeedKmh:
    def test_conversion(self):
        data = make_fix(speed=10.0)  # 10 m/s
        assert abs(data.speed_kmh - 36.0) < 0.001

    def test_zero(self):
        data = make_fix(speed=0.0)
        assert data.speed_kmh == 0.0

    def test_none(self):
        data = make_fix(speed=None)
        assert data.speed_kmh is None


# ------------------------------------------------------------------
# Serialization
# ------------------------------------------------------------------

class TestSerialization:
    def test_to_dict_has_all_fields(self):
        data = make_fix()
        d = data.to_dict()
        for field in ("latitude", "longitude", "altitude", "speed",
                      "heading", "fix_quality", "num_satellites", "hdop", "timestamp"):
            assert field in d

    def test_to_dict_timestamp_is_iso_string(self):
        data = make_fix()
        d = data.to_dict()
        assert isinstance(d["timestamp"], str)
        datetime.fromisoformat(d["timestamp"])  # must not raise

    def test_to_dict_fix_quality_is_int(self):
        data = make_fix()
        d = data.to_dict()
        assert isinstance(d["fix_quality"], int)
        assert d["fix_quality"] == 1

    def test_to_json_is_valid_json(self):
        data = make_fix()
        parsed = json.loads(data.to_json())
        assert parsed["latitude"] == pytest.approx(39.972672)

    def test_roundtrip_dict(self):
        original = make_fix()
        restored = GPSData.from_dict(original.to_dict())
        assert restored.latitude == pytest.approx(original.latitude)
        assert restored.longitude == pytest.approx(original.longitude)
        assert restored.fix_quality == original.fix_quality
        assert restored.timestamp == original.timestamp

    def test_roundtrip_json(self):
        original = make_fix()
        restored = GPSData.from_json(original.to_json())
        assert restored.altitude == pytest.approx(original.altitude)
        assert restored.num_satellites == original.num_satellites

    def test_from_dict_fix_quality_converted(self):
        data = make_fix()
        d = data.to_dict()
        restored = GPSData.from_dict(d)
        assert isinstance(restored.fix_quality, FixQuality)

    def test_roundtrip_no_fix(self):
        original = GPSData(timestamp=TS)
        restored = GPSData.from_json(original.to_json())
        assert restored.has_fix is False
        assert restored.latitude is None


# ------------------------------------------------------------------
# __repr__
# ------------------------------------------------------------------

class TestRepr:
    def test_repr_with_fix(self):
        data = make_fix()
        r = repr(data)
        assert "lat=" in r
        assert "lon=" in r
        assert "sats=7" in r

    def test_repr_no_fix(self):
        data = GPSData(timestamp=TS)
        assert "NO FIX" in repr(data)


# ------------------------------------------------------------------
# Southern / Western hemisphere
# ------------------------------------------------------------------

class TestHemisphere:
    def test_southern_latitude_negative(self):
        data = make_fix(latitude=-33.8688)
        assert data.latitude < 0

    def test_western_longitude_negative(self):
        data = make_fix(longitude=-73.5673)
        assert data.longitude < 0

    def test_roundtrip_south_west(self):
        original = make_fix(latitude=-33.8688, longitude=-73.5673)
        restored = GPSData.from_json(original.to_json())
        assert restored.latitude == pytest.approx(-33.8688)
        assert restored.longitude == pytest.approx(-73.5673)



"""GPS data models."""
from __future__ import annotations

import json
from dataclasses import dataclass, asdict
from datetime import datetime, timezone
from enum import IntEnum


class FixQuality(IntEnum):
    NO_FIX = 0
    GPS = 1
    DGPS = 2
    PPS = 3
    RTK = 4
    FLOAT_RTK = 5
    ESTIMATED = 6
    MANUAL = 7
    SIMULATION = 8


@dataclass
class GPSData:
    """Represents a single GPS measurement."""

    timestamp: datetime
    latitude: float | None = None
    longitude: float | None = None
    altitude: float | None = None        # metres (MSL)
    speed: float | None = None           # m/s
    heading: float | None = None         # degrees (True North)
    fix_quality: FixQuality = FixQuality.NO_FIX
    num_satellites: int = 0
    hdop: float | None = None            # Horizontal dilution of precision
    raw_sentence: str = ""

    # ------------------------------------------------------------------
    # Helpers
    # ------------------------------------------------------------------

    @property
    def has_fix(self) -> bool:
        return self.fix_quality > FixQuality.NO_FIX and self.latitude is not None

    @property
    def speed_kmh(self) -> float | None:
        return self.speed * 3.6 if self.speed is not None else None

    def to_dict(self) -> dict:
        d = asdict(self)
        d["timestamp"] = self.timestamp.isoformat()
        d["fix_quality"] = int(self.fix_quality)
        return d

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_dict(cls, d: dict) -> "GPSData":
        d = dict(d)
        d["timestamp"] = datetime.fromisoformat(d["timestamp"])
        d["fix_quality"] = FixQuality(d["fix_quality"])
        return cls(**d)

    @classmethod
    def from_json(cls, s: str) -> "GPSData":
        return cls.from_dict(json.loads(s))

    def __repr__(self) -> str:
        if self.has_fix:
            return (
                f"GPSData(lat={self.latitude:.6f}, lon={self.longitude:.6f}, "
                f"alt={self.altitude}m, speed={self.speed_kmh:.1f}km/h, "
                f"sats={self.num_satellites})"
            )
        return f"GPSData(NO FIX, sats={self.num_satellites})"
